import { z } from "zod";
import { JitsuFunction } from "@jitsu/protocols/functions";
import { AnalyticsServerEvent } from "@jitsu/protocols/analytics";
import { PostHog } from "posthog-node";
import { HttpRequest, MixpanelCredentials, specialProperties } from "./mixpanel-destination";
import { randomId } from "juava";
import { getEventCustomProperties } from "./lib";
import { parseUserAgent } from "./lib/browser";

const DEFAULT_HOST = "https://app.posthog.com";
export const PosthogDestinationConfig = z.object({
  key: z.string(),
  host: z.string().optional().default(DEFAULT_HOST),
  enableAnonymousUserProfiles: z.boolean().optional().default(false),
  //  sendIdentifyEvents: z.boolean().optional().default(false),
});

export const PosthogDestinationConfigUi: Partial<Record<keyof PosthogDestinationConfig, { documentation: string }>> = {
  enableAnonymousUserProfiles: {
    documentation: "If enabled, anonymous users will be tracked in Mixpanel",
  },
  // sendIdentifyEvents: {
  //   documentation:
  //     "If enabled, any identify() call will send an Identify event to Mixpanel in addition to the profile update",
  // },
};

export type PosthogDestinationConfig = z.infer<typeof PosthogDestinationConfig>;

function setProfileMessage(distinctId: string, event: AnalyticsServerEvent, opts: MixpanelCredentials): HttpRequest {
  const props = event.traits || event.context?.traits || {};
  specialProperties.forEach(prop => {
    if (props[prop]) {
      props[`$${prop}`] = props[prop];
      delete props[prop];
    }
  });
  const setPayload = {
    $token: opts.projectToken,
    $distinct_id: distinctId,
    $ip: event.context?.ip || event.request_ip,
    $set: props,
  };

  return {
    id: randomId(),
    url: "https://api.mixpanel.com/engage?verbose=1#profile-set",
    headers: {
      "Content-type": "application/json",
      Accept: "text-plain",
    },
    payload: [setPayload],
  };
}

function getHostFromUrl(url: string | undefined): string | undefined {
  if (!url) {
    return undefined;
  }
  try {
    return new URL(url).hostname;
  } catch (e) {
    return undefined;
  }
}

function getPathFromUrl(url: string | undefined): string | undefined {
  if (!url) {
    return undefined;
  }
  try {
    return new URL(url).pathname;
  } catch (e) {
    return undefined;
  }
}

function getEventProperties(event: AnalyticsServerEvent) {
  //see https://github.com/PostHog/posthog-js-lite/blob/master/posthog-web/src/context.ts
  const browser = event.context?.userAgent
    ? parseUserAgent(event.context?.userAgent, event.context?.userAgentVendor)
    : undefined;
  return {
    $referrer: event.context?.page?.referrer,
    $referring_domain: event.context?.page?.referring_domain || getHostFromUrl(event.context?.page?.referrer),
    $current_url: event.context?.page?.url,
    $host: event.context?.page?.host || getHostFromUrl(event.context?.page?.url),
    $pathname: event.context?.page?.path || getPathFromUrl(event.context?.page?.url),

    $browser: browser?.name,
    $device: browser?.deviceType,
    $os: browser?.os,
    $browser_version: browser?.browserVersion,

    //implement when it's implemented on a client, doesn't seem like a very important data points
    $screen_dpi: event.context?.screen?.density,
    $screen_height: event.context?.screen?.height,
    $screen_width: event.context?.screen?.width,
    ...getEventCustomProperties(event, {
      exclude: obj => {
        delete obj.referer;
        delete obj.referring_domain;
        delete obj.url;
        delete obj.host;
        delete obj.path;
        delete obj.screen;
        delete obj.page;
        delete obj.width;
        delete obj.height;
      },
    }),
  };
}

const PosthogDestination: JitsuFunction<AnalyticsServerEvent, PosthogDestinationConfig> = async (
  event,
  { props, fetch, log }
) => {
  const client = new PostHog(props.key, { host: props.host || DEFAULT_HOST, fetch: fetch });
  try {
    if (event.type === "identify") {
      if (!event.userId) {
        const distinctId = event.anonymousId || event.traits?.email;
        if (!distinctId) {
          log.info(`No distinct id found for event ${JSON.stringify(event)}`);
        } else if (props.enableAnonymousUserProfiles) {
          client.identify({
            distinctId: distinctId as string,
            properties: { $anon_distinct_id: event.anonymousId || undefined, ...event.traits },
          });
        }
      } else {
        client.identify({
          distinctId: event.userId as string,
          properties: { $anon_distinct_id: event.anonymousId || undefined, ...event.traits },
        });
        if (event.anonymousId) {
          client.alias({ distinctId: event.userId as string, alias: event.anonymousId });
        }
        if (event.traits?.email) {
          client.alias({ distinctId: event.userId as string, alias: event.traits.email as string });
        }
      }
      // if (props.sendIdentifyEvents) {
      //   const distinctId = event.userId || (event.traits?.email as string) || event.anonymousId;
      //   // if (distinctId) {
      //   //   client.capture({
      //   //     distinctId: distinctId as string,
      //   //     event: "Identify",
      //   //     properties: getEventProperties(event),
      //   //   });
      //   // }
      // }
    } else if (event.type === "group") {
      //messages.push(groupMessage(event, opts));
    } else if (event.type === "track") {
      const distinctId = event.userId || event.anonymousId || (event.traits?.email as string);
      if (!distinctId) {
        log.info(`No distinct id found for event ${JSON.stringify(event)}`);
      } else {
        client.capture({
          distinctId: distinctId as string,
          event: event.event || event.name || "Unknown Event",
          properties: getEventProperties(event),
        });
      }
    } else if (event.type === "page") {
      const distinctId = event.userId || event.anonymousId || (event.traits?.email as string);
      if (!distinctId) {
        log.info(`No distinct id found for Page View event ${JSON.stringify(event)}`);
      } else {
        if (event.userId || props.enableAnonymousUserProfiles) {
          client.capture({
            distinctId: distinctId as string,
            event: "$pageview",
            properties: getEventProperties(event),
          });
        }
      }
    }
  } finally {
    await client.shutdownAsync();
  }
};

export default PosthogDestination;
